@@ -14,34 +14,10 @@ typedef unsigned int u32;
 #ifdef _MSC_VER
 typedef signed __int64 s64;
 typedef unsigned __int64 u64;
-/* #define INLINE _inline */
 #else
 typedef long long s64;
 typedef unsigned long long u64;
-/* #define INLINE inline */
 #endif
-static INLINE u64 gmp2u64(mpz_t src)
-{
-  /* mpz_export is terribly slow */
-  u64 ans = mpz_getlimbn(src, 0);
-#if GMP_LIMB_BITS == 32
-  if (mpz_size(src) >= 2)
-    ans |= (u64)mpz_getlimbn(src, 1) << 32;
-#endif
-  return ans;
-}
-static INLINE void u64_2gmp(u64 src, mpz_t dest)
-{
-#if GMP_LIMB_BITS == 64
-  dest->_mp_d[0] = src;
-  dest->_mp_size = (src ? 1 : 0);
-#else
-  /* mpz_import is terribly slow */
-  mpz_set_ui(dest, (u32)(src >> 32));
-  mpz_mul_2exp(dest, dest, 32);
-  mpz_add_ui(dest, dest, (u32)src);
-#endif
-}
 /******************************************************************************/
 
 #ifndef M_LN2
@@ -415,10 +391,10 @@ Perspective"
     a2 = (a+a) % p;
     alpha = powm_16(a2, (p-5)>>3, p);
     beta = mulm_16(a2, mulm_16(alpha,alpha,p), p);
-    b = mulm_16(alpha, mulm_16(a, (beta ? beta-1 : p-1), p), p);
+    b = mulm_16(alpha, mulm_16(a, (beta ? beta-1 : (u32)p-1), p), p);
     return b;
   }
-  else if ( (p & 16) == 9 ) {
+  else if ( (p & 15) == 9 ) {
     s32 a2, alpha, beta, b, d = 1;
     a2 = (a+a) % p;
     alpha = powm_16(a2, (p-9)>>4, p);
@@ -1103,6 +1079,8 @@ sieve polynomials
       if (++i == POLY_SELECT_BITS_TINY)
         break;
     }
+    if (minus_idx == 0)  /* Something has gone wrong.  Exit. */
+      break;
     plus_idx++;
     minus_idx--;
   }
@@ -1135,7 +1113,7 @@ sieve polynomials
    optimal factor size, with later bitfield values selecting
    prime factors that march away from the optimal size.
 
-   popcount[] gives the number of set bits for each value
+   _popcount[] gives the number of set bits for each value
    of the bitfield, and a_choice[] lists the bitfields
    themselves. A given factorization only uses one of the
    population count sizes from the table; bitfields are
@@ -1145,7 +1123,8 @@ sieve polynomials
    same weight, and there are enough bitfields to generate
    256 polynomials, whether A values contain 3, 4, or 5 primes */
 
-static u8 popcount[] = {
+/* DAJ: Renamed because NetBSD is broken. */
+static u8 _popcount[] = {
        3,     4,     3,     5,     3,     4,
        3,     4,     3,     3,     4,     4,
        3,     4,     5,     4,     5,     4,
@@ -1189,11 +1168,11 @@ Compute the next polynomial A value
   /* choose the next bitfield representing
      primes to use */
 
-  for (i = params->poly_select_idx; i < sizeof(popcount); i++) {
-    if (popcount[i] == num_a_factors)
+  for (i = params->poly_select_idx; i < sizeof(_popcount); i++) {
+    if (_popcount[i] == num_a_factors)
       break;
   }
-  if (i >= sizeof(popcount))
+  if (i >= sizeof(_popcount))
     return -1;
   mask = a_choice[i];
   params->poly_select_idx = i + 1;
