@@ -7,7 +7,8 @@ fi
 
 cp -p ptypes.h standalone/
 cp -p ecpp.[ch] bls75.[ch] aks.[ch] ecm.[ch] prime_iterator.[ch] standalone/
-cp -p gmp_main.[ch] factor.[ch] squfof126.[ch] pbrent63.[ch] standalone/
+cp -p gmp_main.[ch] real.[ch] standalone/
+cp -p factor.[ch] squfof126.[ch] pbrent63.[ch] tinyqs.[ch] standalone/
 cp -p utility.[ch] isaac.[ch] random_prime.[ch] standalone/
 cp -p primality.[ch] standalone/
 cp -p xt/expr.[ch] xt/expr-impl.h standalone/
@@ -44,7 +45,8 @@ CFLAGS = -O3 -g -Wall $(DEFINES)
 LIBS = -lgmp -lm
 
 OBJ = ecpp.o bls75.o aks.o primality.o ecm.o prime_iterator.o gmp_main.o \
-      isaac.o random_prime.o squfof126.o pbrent63.o factor.o utility.o expr.o
+      factor.o squfof126.o pbrent63.o tinyqs.o \
+      real.o isaac.o random_prime.o utility.o expr.o
 HEADERS = ptypes.h class_poly_data.h
 
 .PHONY: default all clean
